﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows.Forms;
using System.Diagnostics;
using System.Threading;

namespace LoopThreading
{
    public partial class MainForm : Form
    {
        const int OuterLoops = 500;
        const int InnerLoops = 500;

        TestOperationContext multiThreadedContext;
        readonly Stopwatch stopWatch = new Stopwatch();

        public MainForm()
        {
            InitializeComponent();
        }

        void ByeButton_Click(object sender, EventArgs e)
        {
            Close();
        }

        void Test1Button_Click(object sender, EventArgs e)
        {
            IEnumerable<DataSample> testData = GenerateData();
            var rnd = new Random();

            LogTestStarted("Single-Threaded");

            for (int i = 0; i < OuterLoops; i++)
                TestOperation(new TestOperationContext { RandomSeed = rnd.Next(), Data = testData });

            LogTestCompleted();
        }

        void Test2Button_Click(object sender, EventArgs e)
        {
            IEnumerable<DataSample> testData = GenerateData();
            var rnd = new Random();

            LogTestStarted("Multi-Threaded, ParameterizedThreadStart");

            for (int i = 0; i < OuterLoops; i++)
            {
                var thread = new Thread(ObjectContextTestOperation);
                thread.Start(new TestOperationContext { RandomSeed = rnd.Next(), Data = testData.Skip(InnerLoops / 2) });
                TestOperation(new TestOperationContext { RandomSeed = rnd.Next(), Data = testData.Take(InnerLoops / 2) });
                thread.Join();
            }

            LogTestCompleted();
        }

        void Test3Button_Click(object sender, EventArgs e)
        {
            IEnumerable<DataSample> testData = GenerateData();
            var rnd = new Random();

            LogTestStarted("Multi-Threaded, Class-Local Context");

            for (int i = 0; i < OuterLoops; i++)
            {
                multiThreadedContext = new TestOperationContext { RandomSeed = rnd.Next(), Data = testData.Skip(InnerLoops / 2) };
                var thread = new Thread(ContextlessTestOperation);
                thread.Start();
                TestOperation(new TestOperationContext { RandomSeed = rnd.Next(), Data = testData.Take(InnerLoops / 2) });
                thread.Join();
            }

            LogTestCompleted();
        }

        void Test4Button_Click(object sender, EventArgs e)
        {
            IEnumerable<DataSample> testData = GenerateData();
            var rnd = new Random();

            LogTestStarted("Multi-Threaded, Generic Context & Delegate");

            for (int i = 0; i < OuterLoops; i++)
            {
                var thread = new GenericContextThread<TestOperationContext>(TestOperation);
                thread.Start(new TestOperationContext { RandomSeed = rnd.Next(), Data = testData.Skip(InnerLoops / 2) });
                TestOperation(new TestOperationContext { RandomSeed = rnd.Next(), Data = testData.Take(InnerLoops / 2) });
                thread.Join();
            }

            LogTestCompleted();
        }

        void Test5Button_Click(object sender, EventArgs e)
        {
            IEnumerable<DataSample> testData = GenerateData();
            var rnd = new Random();

            using (var thread = new PersistentThread<TestOperationContext>(TestOperation))
            {
                LogTestStarted("Multi-Threaded, PersistentThread (single kept-alive thread w/ generic context & delegate)");

                for (int i = 0; i < OuterLoops; i++)
                {
                    thread.Start(new TestOperationContext { RandomSeed = rnd.Next(), Data = testData.Skip(InnerLoops / 2) });
                    TestOperation(new TestOperationContext { RandomSeed = rnd.Next(), Data = testData.Take(InnerLoops / 2) });
                    thread.Join();
                }

                LogTestCompleted();
            }
        }

        void Test6Button_Click(object sender, EventArgs e)
        {
            IEnumerable<DataSample> testData = GenerateData();
            var rnd = new Random();

            LogTestStarted("Task Parallel Library");

            for (int i = 0; i < OuterLoops; i++)
                Parallel.ForEach(testData, sample => { sample.Number = SlowFunctions.Bessel(sample.Number / rnd.Next()); });

            LogTestCompleted();
        }

        void Test7Button_Click(object sender, EventArgs e)
        {
            IEnumerable<DataSample> testData = GenerateData();
            var rnd = new Random();

            const int threadCount = 3;
            const int dataPerThread = InnerLoops / (threadCount + 1);

            var threads = new PersistentThread<TestOperationContext>[threadCount];
            for (int i = 0; i < threadCount; i++)
                threads[i] = new PersistentThread<TestOperationContext>(TestOperation);

            LogTestStarted("Multi-Threaded, 3x PersistentThread");

            for (int i = 0; i < OuterLoops; i++)
            {
                for (int j = 0; j < threadCount; j++)
                    threads[j].Start(new TestOperationContext { RandomSeed = rnd.Next(), Data = testData.Skip((j + 1) * dataPerThread).Take(dataPerThread) });

                TestOperation(new TestOperationContext { RandomSeed = rnd.Next(), Data = testData.Take(dataPerThread) });

                for (int j = 0; j < threadCount; j++)
                    threads[j].Join();
            }

            for (int i = 0; i < threadCount; i++)
                threads[i].Dispose();

            LogTestCompleted();
        }

        void LogTestStarted(string description)
        {
            ConsoleText.Text += string.Format("Test '{0}' Started... ", description);
            Application.DoEvents();

            stopWatch.Start();
        }

        void LogTestCompleted()
        {
            stopWatch.Stop();

            ConsoleText.Text += "Completed.\r\n"; 
            ConsoleText.Text += string.Format("Time Elapsed : {0}\r\n\r\n", stopWatch.Elapsed);

            stopWatch.Reset();
        }

        void ClearButton_Click(object sender, EventArgs e)
        {
            ConsoleText.Clear();
        }

        static IEnumerable<DataSample> GenerateData()
        {
            return Enumerable.Range(0, InnerLoops).Select(x => new DataSample { Number = x });
        }

        static void ObjectContextTestOperation(object objectContext)
        {
            var context = (TestOperationContext)objectContext;
            foreach (DataSample sample in context.Data)
                sample.Number = SlowFunctions.Bessel(sample.Number / context.RandomSeed);
        }

        void ContextlessTestOperation()
        {
            foreach (DataSample sample in multiThreadedContext.Data)
                sample.Number = SlowFunctions.Bessel(sample.Number / multiThreadedContext.RandomSeed);
        }

        static void TestOperation(TestOperationContext context)
        {
            foreach (DataSample sample in context.Data)
                sample.Number = SlowFunctions.Bessel(sample.Number / context.RandomSeed);
        }
    }
}
