﻿using System;
using System.Threading;

namespace LoopThreading
{
    public class PersistentThread<ContextType> : IDisposable
    {
        readonly Thread thread;
        readonly ManualResetEvent startEvent, joinEvent;
        readonly Action<ContextType> method;

        ContextType cachedContext;
        bool started;
        bool disposed;

        public PersistentThread(Action<ContextType> method)
        {
            startEvent = new ManualResetEvent(false);
            joinEvent = new ManualResetEvent(false);

            this.method = method;
            thread = new Thread(DoWork);

            thread.Start();
        }

        public void Start(ContextType context)
        {
            if (disposed)
                throw new ObjectDisposedException("PersistentThread");
            if (started)
                throw new InvalidOperationException("Thread is already started");

            started = true;
            cachedContext = context;

            startEvent.Set();
        }

        public void Join()
        {
            if (disposed)
                throw new ObjectDisposedException("PersistentThread");
            if (!started)
                throw new InvalidOperationException("Thread is idle");

            joinEvent.WaitOne();

            started = false;
        }

        void DoWork()
        {
            startEvent.WaitOne();

            while (!disposed)
            {
                method(cachedContext);
                joinEvent.Set();
                startEvent.WaitOne();
            }
        }

        public void Dispose()
        {
            if (!disposed)
                GC.SuppressFinalize(this);

            DisposeInternal();
        }

        void DisposeInternal()
        {
            if (!disposed)
            {
                disposed = true;
                startEvent.Set();
            }
        }

        ~PersistentThread()
        {
            DisposeInternal();
        }
    }
}
